from ._decision_tree import RegressionTree, ClassificationTree
from machine_learning._templates import RegressorTemplate
from machine_learning._templates import ClassifierTemplate


class RandomForestRegressor(RegressorTemplate):
    pass


class RandomForestClassifier(ClassifierTemplate):
    pass
